@@ -1,5 +1,4 @@
-#ifndef __SVGLITE_UTILS__
-#define __SVGLITE_UTILS__
+#pragma once
 
 #include <limits>
 #include <cmath>
@@ -12,7 +11,7 @@ inline static double dbl_format(double x) {
     return x;
 }
 
-bool iequals(const std::string& a, const std::string& b) {
+inline bool iequals(const std::string& a, const std::string& b) {
   unsigned int sz = a.size();
   if (b.size() != sz) {
     return false;
@@ -24,5 +23,3 @@ bool iequals(const std::string& a, const std::string& b) {
   }
   return true;
 }
-
-#endif
